function[final_matrices, final_bgs] = em(seqs, priors, bg_model, motif_model, pcounts, num_motifs, update_bg_model)

log_priors = log(priors);
window = size(motif_model, 2);

expectation_array = zeros(size(seqs, 1), size(seqs, 2) - window);
prev_total_likelihood = 0.0;
log_mask_array = zeros(size(expectation_array));

min_iterations = 10; %% min # of E/M iterations
max_iterations = 100;   %% max # of E/M iterations
run_cutoff = 5;  %% the # of "slowing" increases in the total log-likelihood seen, before we stop iterating.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% We will travel through this loop once for 
%% every motif we set out to discover.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for m =1:num_motifs
	
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%% Do some initialization for each motif.
	%% We keep track of, among other things, the difference
	%% between successive total-data log-likelihoods 
	%% (diff) and successive diff's (diff_diff).
	%% We stop iterating when the diff_diff (which is a *rough*
	%% approximation to the 2nd derivative of "total-log-likelihood
	%% vs. time") stays negative for a "sufficient" number
	%% (dictated by 'run_cutoff') iterations.
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	log_motif_model = log(motif_model);
	log_bg_model = log(bg_model);
	i = 0;
	diff = 0.0;
	diff_diff = 0.0;
	run_direction = 1;
	run_size = 0;
	prev_diff = 0.0;
	continue_iterating = 1;

	fprintf(1, '\nMotif %d: --> %s\n', m, matrix2string(motif_model));

	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%% This inner-loop is the iterating E/M steps 
	%% of EM.  It is phrased as a "while" loop because 
	%% of the (relatively) complicated stopping 
	%% conditions (see below).
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	while continue_iterating == 1

		total_likelihood = 0.0;

		%% --------------------------------------------------
		%% This is the core of the EM code: for each sequence,
		%% calculate it's expectation, and updated the tracked
		%% total likelihood... Then, updated the FG and BG
		%% models according to this expectation.
		%% --------------------------------------------------
		for j = 1:size(seqs, 1)

			%% ***** E-step *****
			[likelihood, expectation_array(j, :)] = expectation(j, seqs, log_priors, log_mask_array, log_motif_model, log_bg_model);

			total_likelihood = total_likelihood + likelihood;

		end

		%% ***** M-step ******
		[updated_motif_model, updated_bg_model] = maximization(seqs, expectation_array, pcounts);

		%% we allow the user to specify whether the background is updated at each iteration or 
		%% not, using the "update_bg_model" parameter (1 == update, 0 = don't).
		if update_bg_model == 1
			log_bg_model = log(updated_bg_model);
		else
			log_bg_model = log(bg_model);
		end
		log_motif_model = log(updated_motif_model);

		%% --------------------------------------------------
		%% bookkeeping and output, at each iteration: 
		%% update the length of the current run (is our total likelihood diff getting bigger
		%% or smaller?) and output all the current statistics.
		%% --------------------------------------------------
		fprintf(1, '%d: \t', i);
		fprintf(1, '%s\t', matrix2string(updated_motif_model));
		fprintf(1, '%f\t ', total_likelihood);
		if i > 0 
			diff = total_likelihood - prev_likelihood;
			if i > 1
				%% In here, we're keeping track of the sign of the 
				%% 2nd-derivative approximation.
				diff_diff = diff - prev_diff;
				if diff_diff > 0.0
					fprintf(1, '+');
					if run_direction == -1
						run_direction = 1;
						run_size = 0;
					end
				else
					fprintf(1, '-');
					if run_direction == 1
						run_direction = -1;
						run_size = 0;
					end
				end
				run_size = run_size + 1;
			else
				fprintf(1, ' ')
			end
			fprintf(1, ' (%f)', diff);
		end
		fprintf(1, '\n');

		%% --------------------------------------------------
		%% Update our parameters for the next iteration.
		%% --------------------------------------------------
		prev_likelihood = total_likelihood;
		prev_diff = diff;
		i = i + 1;

		%% --------------------------------------------------
		%% finally, perform our simple convergence test: has our likelihood difference 
		%% been getting consistently smaller?
		%% --------------------------------------------------
		if i > min_iterations 
			if i > max_iterations || (run_direction == -1 && run_size > run_cutoff)
				continue_iterating = 0;
			end
		end
	end

	%% --------------------------------------------------
	%% Some final output, and storage of the results.
	%% --------------------------------------------------
	final_matrices(m, :, :) = updated_motif_model;
	final_bgs(m, :) = updated_bg_model;
	[max_offsets,log_mask_array] = mask_log_priors(log_mask_array, log(expectation_array), window);
	max_words = find_maximum_words(log(expectation_array), seqs, window);
	fprintf(1, '\n--- Maximum Offsets:\n');
	for w=1:size(max_words, 1)
		fprintf(1, '%d: \t(%d) \t%s\n', w, max_offsets(w), max_words(w, :));
	end
	if update_bg_model == 1
		display(updated_bg_model);
	end
	view_matrix(final_matrices, m);

end



